/*_agicall_api*/
var ai_err_pending = -1;
var ai_err_no = 0;
var ai_err_ignore = 1;
var ai_err_not_open = 80001;
var ai_err_invalid_param = 80002;
var ai_err_disconnect = 80003;
var ai_err_url = 80004;
var ai_all_account = "_all_account_";//匿名账号绑定全部通道

function _agicall_base() {
    var m_url = '';
    var m_retry_delay = 3000;//0表示不重试
    var m_timeout = 5000;//正常tcp连接请求超时N毫秒,可能业务层需要处理命令,操作数据库延迟等,所以加大默认超时时间到5秒

    this.get_url = function () { return m_url; }
    this.set_url = function (url) { m_url = url; }
    this.set_retry_delay = function (retry_delay) { m_retry_delay = retry_delay; }
    this.get_retry_delay = function () { return m_retry_delay; }
    this.set_timeout = function (t) { m_timeout = t; }
    this.get_timeout = function () { return m_timeout; }
}

function _agicall_ws() {
    var m_ws_ = null;
    var m_re_connect = true;//是否重连,手工disconnect断开后就不会重连了
    var m_re_connect_timer = 0;
    var m_connect_timer = 0;
    var m_open_ws_time = 0;//打开时间,结束可以判断连接时间,间隔时间太短的是异常需要等会再连接,避免死循环,一般tcp不会立即返回失败,所以可以在断开后立即重连
    var self = this;

    _agicall_base.call(this);

    this.connect_url = function () {
        if (!m_ws_) {
            return do_connect_ws(self.get_url());
        }
        else return ai_err_no;
    }

    this.disconnect = function () {
        m_re_connect = false;
        stop_re_connect();
        stop_connect_timeout();
        if (!m_ws_) return ai_err_ignore;
        else {
            m_ws_.close();
            return ai_err_no;
        }
    }

    var do_connect_ws = function (url) {
        self.d_logout("->do ws connect:" + url);
        free_ws_();
        m_re_connect = self.get_retry_delay() > 0 ? true : false;
        m_ws_ = new WebSocket(url);
        m_ws_.onopen = self.onopen_ws;
        m_ws_.onmessage = self.onmessage_ws;
        m_ws_.onerror = self.onerror_ws;
        m_ws_.onclose = self.onclose_ws;
        start_connect_timeout();
        return 0;
    }
    this.is_connected = function () { return self.is_open(); }
    //0:CONNECTING/1:open/2:closing/3:closed
    this.is_open = function () {
        return (m_ws_ != null && m_ws_.readyState == 1) ? true : false;
    }

    this.send = function (s_data, timeout, op_key) {
        if (self.is_open()) {
            m_ws_.send(s_data);
            return ai_err_no;
        }
        else {
            return ai_err_not_open;
        }
    }

    var free_ws_ = function () {
        if (m_ws_) {
            m_ws_.close();
            m_ws_ = null;
        }
        return ai_err_no;
    }

    var start_connect_timeout = function () {
        stop_connect_timeout();
        if (self.get_timeout() > 0) {
            m_connect_timer = window.setTimeout(function (self_obj) { self_obj.on_connect_timeout() }, self.get_timeout(), self);
        }
    };
    var stop_connect_timeout = function () {
        if (m_connect_timer > 0) {
            window.clearTimeout(m_connect_timer);
            m_connect_timer = 0;
        }
    }

    this.on_connect_timeout = function () {
        self.d_logout('connect timeout');
        free_ws_();
    }

    var stop_re_connect_timer = function () {
        if (m_re_connect_timer != 0) {
            window.clearTimeout(m_re_connect_timer);
            m_re_connect_timer = 0;
        }
    }

    var start_re_connect = function () {
        stop_re_connect_timer();
        if (self.get_retry_delay() > 0 && m_re_connect) {
            if (self.get_retry_delay() < 10)//可以认为无延迟,立即连接
            {
                self.connect_url();
            }
            else {
                m_re_connect_timer = window.setTimeout(function (self_obj) { self_obj.on_re_connect_timer() }, self.get_retry_delay(), self);
            }
        }
        else {
            self.d_logout("ignore re connect retry_delay=0");
        }
    }

    this.on_re_connect_timer = function () {
        if (m_re_connect && m_re_connect_timer != 0) {
            stop_re_connect_timer();
            self.connect_url();
        }
        else {
            //self.d_logout("ignore on_re_connect_timer..." + self_obj.m_re_connect + " " + self_obj.m_re_connect_timer);
        }
        return 1;
    }

    this.on_notify_msg = function (js_Obj) {
        var act = js_Obj.data.action;
        if (act == 'bind_acc_timeout') {
            self.d_logout('***will close by wait bind acc timeout***');
            self.on_bind_wait_timeout(js_Obj.data.timeout);
        }
    }

    var stop_re_connect = function () {
        self.d_logout("stop_re_connect:" + m_re_connect);
        m_re_connect = false;//手工操作不用重连
        stop_re_connect_timer();
    }

    this.onopen_ws = function () {
        self.d_logout('ws on open->' + self.get_url());
        m_open_ws_time = new Date();
        stop_connect_timeout();
        self.on_connected(self.get_url());
    }

    this.onerror_ws = function () { // const onerror_WS = () => {               
        self.d_logout('ws on error->' + self.get_url());
        //self.set_unbind_all_account();
        self.on_error(self.get_url());
        free_ws_();
    }

    this.onclose_ws = function () {
        if (m_ws_) {
            self.d_logout('ws on close:' + self.get_url());
            m_ws_ = null;
        }
        self.on_disconnected(self.get_url());
        stop_connect_timeout();
        if (m_re_connect) {
            start_re_connect();
        }
    }

    this.onmessage_ws = function (ws_pkt) {//const onmessage_WS = ws_pkt => {
        if (typeof (ws_pkt.data) == 'string') {
            var js_Obj = JSON.parse(ws_pkt.data);
            self.on_recv_message(js_Obj);
            self.on_recv_end(js_Obj.content.action, 4, 200, 1000);//模拟跟http一样表示接收完毕后处理下一步逻辑           
        }
    }
    this.on_error = function (url) { }
    this.d_logout = function (s) {
        u_logout(s);
    }
}

function _agi_url() {
    var m_proto = 0;//0:未知， 1:http/https  2:ws/wss
    var m_url = '';
    var m_query_evt_tick = 0;
    var m_is_query_evt = false;
    var m_xhr_http = null;
    var m_xhr_ws = null;
    var m_xhr_query_evt = null;
    var self = this;

    this.get_query_evt_tick = function () { return m_query_evt_tick; }
    this.reset_query_evt_tick = function () { m_query_evt_tick = 0; }
    this.get_is_query_evt = function () { return m_is_query_evt; }
    this.get_url = function () { return m_url; }
    this.set_url = function (url) {
        m_url = url;
        var proto = u_url_protocol(url);
        if (u_isEmpty(proto)) return 1;
        else if (proto == 'http' || proto == 'https') {
            m_proto = 1;
            return 0;
        } else if (proto == 'ws' || proto == 'wss') {
            m_proto = 2;
            if (m_xhr_ws) {//断开原来已经创建的连接
                m_xhr_ws.disconnect();
                m_xhr_ws = null;
            }
            //每次都重新new ，ws内部connect 会判断是否是空的,空的才会连接url
            m_xhr_ws = new _agicall_ws();
            m_xhr_ws.set_url(url);
            m_xhr_ws.on_connected = function (url) { self.on_connected(url); }
            m_xhr_ws.on_disconnected = function (url) { self.on_disconnected(url); }
            m_xhr_ws.on_error = function (url) { self.on_error(url); }
            m_xhr_ws.on_recv_message = function (js_obj) { self.on_recv_message(js_obj); }
            m_xhr_ws.on_recv_end = function (op_key, readyState, status, duration) { if (op_key == 'query_evt') { m_is_query_evt = false; } self.on_recv_end(op_key, readyState, status, duration); }
            m_xhr_ws.on_bind_wait_timeout = function (t) { self.on_bind_wait_timeout(t); }
            return ai_err_no;
        }
        else {
            return 2;
        }
    }

    this.connect = function () {
        if (m_proto == 1) {
            self.on_connected(m_url); return 0;
        }
        else if (m_proto == 2) {
            if (m_xhr_ws) {
                return m_xhr_ws.connect_url();
            } else {
                return ai_err_ignore;
            }

        }
        else {
            return ai_err_ignore;
        }
    }

    this.disconnect = function () {
        if (m_xhr_query_evt) m_xhr_query_evt.abort();
        if (m_xhr_ws) m_xhr_ws.disconnect();
        m_is_query_evt = false;
        m_query_evt_tick = 0;
    }

    this.is_open = function () {
        if (m_proto == 1) { return true; }
        else if (m_proto == 2 && m_xhr_ws) {
            return m_xhr_ws.is_open();
        } else return false;
    }
    this.http_req = function (s_data, timeout, op_key) {
        var xhr_http = new xhr_post_json(self.get_url(), timeout, op_key);
        xhr_http.ontimeout = function () { u_logout('_aicall_http xhr timeout:' + self.get_url()); }
        xhr_http.onopen = function () { }
        xhr_http.onmessage = function (s_data) { self.on_recv_message(JSON.parse(s_data)); }
        xhr_http.onerror = function (err) { self.on_error(self.get_url()); }
        xhr_http.onclose = function (op_key, readyState, status, duration) {
            if (op_key == 'query_evt') { m_is_query_evt = false; m_xhr_query_evt = null; }
            self.on_recv_end(op_key, readyState, status, duration); xhr_http = null;
        }
        m_xhr_http = xhr_http.send(s_data);
        if (op_key == 'query_evt') {
            m_xhr_query_evt = m_xhr_http;
        }
        return ai_err_no;
    }

    this.send = function (s_data, timeout, op_key) {
        if (op_key == 'query_evt' && m_is_query_evt)//同时只能一次请求查询事件
        {
            u_logout('**ignore repeat query evt...');
            return ai_err_ignore;
        }
        var ret = ai_err_ignore;
        if (m_proto == 1) {
            ret = self.http_req(s_data, timeout, op_key);
        }
        else if (m_proto == 2) {
            if (m_xhr_ws) {
                ret = m_xhr_ws.send(s_data);
            }
            else {
                u_logout('***send ws is empty');
            }
        }
        else {
            u_logout('***send proto error:' + m_proto);
        }
        if (ret == ai_err_no && op_key == 'query_evt') {
            m_is_query_evt = true;
            m_query_evt_tick = new Date().getTime();
        }
        return ret;
    }

    this.on_recv_message = function (js_obj) { u_logout('on_recv_message'); }
    this.on_connected = function (url) { u_logout('on connected->' + url); }
    this.on_disconnected = function (url) { u_logout('on disconnected->' + url) }
    this.on_error = function (url) { u_logout('on_error'); }
    this.on_recv_end = function (op_key, readyState, status, duration) { u_logout('on_recv_end'); }
    this.on_bind_wait_timeout = function (t) { }
}


function _agicall_api() {
    var m_charset='utf-8';//保留,固定返回utf-8格式
    var m_auth_key = '';//服务器返回的授权key,类似某些系统的token
    var m_pri_key = '';//第一次绑定成功后用服务器指定的账号 设置为 私有秘钥 md5_pwd, 或者增加一个agent参数，用MD5[坐席id+密码作为私钥]
    var m_user_key = u_getutc_ms();//客户端key,加密使用
    var m_act_id = u_getutc_ms();//请求操作递增标记,服务端可以用来判断是否重发的操作
    var m_enable_query_evt = true;//绑定成功后可以投递
    var m_evt_uid = 0;//收到的最后一个事件id,是针对本授权的通道,收到的应该是递增的,不递增的就是丢包了
    var m_evt_eid = 0;//全局事件的id,收到的不一定是递增的
    var m_bind_acc = {};//账号索引map object,不是数组
    var m_unbind_acc = {};//想绑定还未绑定的,等待交换授权成功后可以重试
    var m_agent_data = {};//当前绑定账号的坐席信息,包括坐席id和密码其它可以用户自己扩展，这个密码可以用来当私钥
    var m_timeout = 5000;//正常tcp连接请求超时N毫秒,可能业务层需要处理命令,操作数据库延迟等,所以加大默认超时时间到5秒
    var m_query_evt_timeout = 30000;//发送query_evt 后等待N毫秒超时，实际应用建议30秒
    var m_retry_delay = 3000;//0表示不重试
    var m_retry_timer = 0;//重新授权定时器
    var m_act_times = {};//某个操作正在进行的次数,一般是1次
    var m_pop_evt = true;//ws是否是pop模式,服务器发送后不管是否收到就删除事件
    var m_max_evt_num = 10;//一次最多返回的事件数量
    var m_disconnect = true;//手工断开标记,断开后不在自动重连
    var m_logout = true;//是否日志输出

    var m_agi_url = new _agi_url();

    var self = this;

    this.get_version = function () { return "23.04.20"; }
    this.set_logout = function (b) { m_logout = b; }
    this.set_charset = function(c) { m_charset = c;}
    this.get_logout = function () { return m_logout; }
    this.set_pop_evt = function (b) { m_pop_evt = b; }
    this.get_pop_evt = function () { return m_pop_evt; }
    this.set_max_evt_num = function (num) { m_max_evt_num = num; }
    this.get_max_evt_num = function () { return m_max_evt_num; }
    this.set_url = function (url) { return m_agi_url.set_url(url); }
    this.get_url = function () { return m_agi_url.get_url(); }
    this.set_disconnect = function (b) { m_disconnect = b; }
    this.get_disconnect = function () { return m_disconnect; }
    this.set_retry_delay = function (retry_delay) { m_retry_delay = retry_delay; }
    this.get_retry_delay = function () { return m_retry_delay; }
    this.set_auth_key = function (key) { m_auth_key = key; }
    this.get_auth_key = function () { return m_auth_key; }
    this.set_timeout = function (t) { m_timeout = t; }
    this.get_timeout = function () { return m_timeout; }
    this.get_user_key = function () { return m_user_key; }

    this.set_agent_data = function (json_agent) { m_agent_data = json_agent; }

    //====继承通信接口返回
    m_agi_url.on_connected = function (url) {
        self.set_disconnect(false);
        //已经有的授权号码不请求,绑定时提示auth错误再重新获取
        //强制每次连接后获取auth_key可能意外断开网络后中间突然的事件会丢失，但是程序逻辑简单   
        if (u_isNotEmpty(self.get_auth_key())) {
            self.bind_account_try();
        }
        else {
            self.query_auth_key();
        }
    }
    m_agi_url.on_disconnected = function (url) {
        u_logout('on disconnected->' + url);
        self.set_unbind_all_account();
        self.set_disconnect(true);
    }

    m_agi_url.on_error = function (url) { self.on_url_error(url); }
    m_agi_url.on_recv_end = function (op_key, readyState, status, duration) { self.do_recv_end(op_key, readyState, status, duration); }
    m_agi_url.on_recv_message = function (js_obj) { self.do_recv_message(js_obj); }
    //===
    this.connect = function (url) {
        if (self.set_url(url) != 0) return ai_err_url;
        self.set_disconnect(false);
        return m_agi_url.connect();
    }
    //释放所有资源,需要重新连接绑定
    this.disconnect = function () {
        self.remove_auth_key();
        m_bind_acc = {};
        m_unbind_acc = {};
        self.set_disconnect(true);
        self.stop_query_event();
        self.stop_query_auth_timer();
        m_agi_url.disconnect();
        return true;
    }

    //使用深度拷贝,避免返回引用被外部修改
    var copy_acc_array = function (acc_array) {
        var ret_array = new Array();
        for (var v_acc in acc_array) {
            ret_array.push(v_acc);
        }
        return ret_array;
    }

    this.get_unbind_account = function () {
        return copy_acc_array(m_unbind_acc);
    }
    this.get_bind_account = function () {
        return copy_acc_array(m_bind_acc);
    }

    this.is_bind_account = function (acc) {
        if (m_bind_acc[acc]) { return true; }
        else return false;
    }

    //数组形式提供账号/密码/class,这样允许一次多个
    this.bind_account_array = function (acc_array) {
        if (!Array.isArray(acc_array)) return ai_err_ignore;
        else {
            var json_account = new Array();
            for (n = 0; n < acc_array.length; n++) {
                var v_acc_cfg = acc_array[n];
                struct_acc_unbind(v_acc_cfg['account'], v_acc_cfg['pwd'], v_acc_cfg['pwd_type'], v_acc_cfg['class'], json_account);
            }
            if (!u_isNotEmpty(m_auth_key)) {
                self.query_auth_key();
                return ai_err_pending;
            }
            else return do_bind_json_param(json_account);
        }
    }
    //转换到多个数组支持模式调用
    this.bind_account_cls = function (acc, pwd, pwd_type, cls) {
        var acc_array = new Array();
        var acc_cfg = {
            'charset': m_charset,
            'account': acc,
            'class': cls,
            'pwd': pwd,
            'pwd_type': pwd_type
        }
        acc_array.push(acc_cfg);
        return self.bind_account_array(acc_array);
    }

    this.bind_account = function (acc, pwd, pwd_type) {
        if (u_isEmpty(acc)) acc = ai_all_account;
        return self.bind_account_cls(acc, pwd, pwd_type, '');
    }
    //内部使用单个对象发送,不用特意循环构造数组,保留
    var bind_account_single_cls = function (acc, pwd, cls) {
        var json_account = new Array();
        self.struct_acc_param(acc, pwd, cls, json_account);
        return do_bind_json_param(json_account);
    }
    var get_md5_pwd = function (id, pwd, pwd_type) {
        if (u_isNotEmpty(pwd)) {
            if (pwd_type == 'md5') return pwd;
            else return u_md5(id + '_' + pwd + '_agicall.com');
        }
        else return '';
    }
    var get_auth_md5_pwd = function (md5_pwd) {
        if (u_isNotEmpty(md5_pwd)) {
            return u_md5(md5_pwd + '_' + m_auth_key + '_' + m_user_key);
        }
        else return '';
    }
    this.bind_account_try = function () {
        var json_account = new Array();
        for (var v_acc in m_unbind_acc) {
            var json_acc = {                
                'account': v_acc,
                'class': m_unbind_acc[v_acc]['class'],
                'pwd': get_auth_md5_pwd(m_unbind_acc[v_acc]['md5_pwd'])
            }
            json_account.push(json_acc);
        }
        return do_bind_action_account('bind', json_account);
    }

    var do_bind_action_account = function (bd, json_account) {
        if (json_account.length > 0) {
            var v_agent = {};
            //重新构造md5_pwd
            m_agent_data['md5_pwd'] = get_md5_pwd(m_agent_data['id'], m_agent_data['pwd'], m_agent_data['pwd_type']);
            for (var v_key in m_agent_data) {
                if (v_key == 'pwd') { }
                else if (v_key == 'md5_pwd') {
                    v_agent['pwd'] = get_auth_md5_pwd(m_agent_data['md5_pwd']);
                }
                else if (v_key != 'pwd_type') {
                    v_agent[v_key] = m_agent_data[v_key];
                }
            }
            var json_param =
            {
                'charset': m_charset,
                'acc_list': json_account,
                'agent': v_agent
            }
            return do_action_param(bd, json_param);
        }
        else return ai_err_no;
    }
    //解绑指定账号
    this.unbind_account = function (acc) {
        if (u_isEmpty(acc))  return this.unbind_all_account();
        else{
            //if (u_isEmpty(acc)) acc = ai_all_account;        
            if (!m_bind_acc[acc]) return ai_err_ignore;//已经不存在返回不操作
    
            if (self.is_open()) {
                var json_account = new Array();
                var json_acc = { 'account': acc }
                json_account.push(json_acc);
                return do_bind_action_account('unbind', json_account);
            }
            delete m_bind_acc[acc];//手工解绑可以删除掉
            return ai_err_no;
        }
    }
    //解绑全部,如果服务器已经断开的是不是就可以直接删除掉,不会再回复成功了
    this.unbind_all_account = function () {
        if (self.is_open() && Object.keys(m_bind_acc).length > 0) {
            do_logout('unbind all account');
            self.on_unbind_all_account(m_bind_acc);
            var json_account = new Array();
            for (var v_acc in m_bind_acc) {
                var json_acc = { 'account': v_acc }
                json_account.push(json_acc);
            }
            do_bind_action_account('unbind', json_account);
        }
        for (var v_acc in m_bind_acc) {
            delete m_bind_acc[v_acc];
        }
        return ai_err_no;
    }

    //只是把已经绑定的修改成未绑定的状态,下次连接成功自动重新绑定
    this.set_unbind_all_account = function () {
        if (Object.keys(m_bind_acc).length > 0) {
            do_logout('set unbind all account');
            for (var v_acc in m_bind_acc) {
                m_unbind_acc[v_acc] = m_bind_acc[v_acc];
            }
        }
        self.on_set_unbind_all_account(m_bind_acc);
        for (var v_acc in m_bind_acc) {
            delete m_bind_acc[v_acc];
        }
        return ai_err_no;
    }

    this.remove_auth_key = function () {
        if (u_isEmpty(m_auth_key)) return ai_err_no;
        var json_param = {}
        this.set_auth_key('');
        return do_action_param('unauth', json_param);
    }

    this.query_auth_key = function () {
        var json_param = {}
        this.set_auth_key('');//先强制设置为空
        return do_action_param('auth', json_param);
    }

    var do_bind_json_param = function (json_account) {
        if (u_isNotEmpty(m_auth_key) && m_agi_url.is_open()) {
            return do_bind_action_account('bind', json_account);
        }
        else {
            return ai_err_pending;
        }
    }
    var struct_acc_unbind = function (acc, pwd, pwd_type, cls, json_param) {
        var md5_pwd = get_md5_pwd(acc, pwd, pwd_type);
        var json_acc = {
            'account': acc,
            'class': cls,
            'pwd': get_auth_md5_pwd(md5_pwd)//先设置绑定时还没有交换auth_key密码是不对的,try里需要自行重新计算      
        }
        m_unbind_acc[acc] = {
            'class': cls,
            'pwd': pwd,
            'md5_pwd': md5_pwd
        };
        json_param.push(json_acc);
    }

    var new_action = function (act) {
        var json_action = {
            'action': act,
            'act_id': m_act_id
        }
        var json_auth = {
            'auth_key': m_auth_key,
            'auth_crc': u_md5(m_agent_data['md5_pwd'] + '_' + m_auth_key + '_' + m_user_key + '_' + m_act_id)//后期可以考虑再加锁私钥  + m_pri_key
        }
        if (act == 'auth') {//只需要再授权交换key时带上user_key,其它操作可以不带上
            json_auth['user_key'] = m_user_key;
            json_auth['pop_evt'] = m_pop_evt;
            json_auth['max_evt'] = m_max_evt_num;
        }
        json_action['auth'] = json_auth;//可以用来身份鉴别
        m_act_id++;
        return json_action;
    }

    var do_action_param = function (act, json_param) {
        return do_action_param_timeout(act, json_param, self.get_timeout());
    }

    var do_action_param_timeout = function (act, json_param, timeout) {
        var json_action = new_action(act);
        json_action['param'] = json_param;
        add_act_times(act);
        return self.send(JSON.stringify(json_action), timeout, act);
    }

    //需要事件让客户端主动请求,确保不丢id. 也是确认收到事件的方式,确认本id就可以发送下一个id了
    var do_query_event = function () {
        var obj_id = {
            'uid': m_evt_uid,
            'eid': m_evt_eid
        }
        var json_param = {
            'oid': obj_id,
            'timeout': m_query_evt_timeout
        }
        do_logout('do query evt_id:' + m_evt_eid + ' act_id:' + m_act_id);
        return do_action_param_timeout('query_evt', json_param, m_query_evt_timeout + 10000);//客户端延迟10秒超时,目的让服务器能主动超时
    }

    var add_act_times = function (act) {
        if (parseInt(m_act_times[act]) > 0) {
            m_act_times[act] = m_act_times[act] + 1;
        }
        else {
            m_act_times[act] = 1;
        }
        return m_act_times[act];
    }
    var inc_act_times = function (act) {
        if (parseInt(m_act_times[act]) > 0) {
            m_act_times[act] = m_act_times[act] - 1;
        }
        else {
            m_act_times[act] = 0;
        }
        return m_act_times[act];
    }

    var start_query_event = function () {//允许启动投递并且还没启动投递,并且有绑定成功的才可以
        if (m_enable_query_evt && Object.keys(m_bind_acc).length > 0) {
            do_query_event();
        }
        else {
            do_logout('enable:' + m_enable_query_evt + ' acc:' + Object.keys(m_bind_acc).length)
        }
    }

    this.stop_query_event = function () {
        m_enable_query_evt = false;
    }
    //---
    this.do_recv_end = function (op_key, readyState, status, duration) {
        inc_act_times(op_key);
        if (self.get_disconnect()) { return 0; }
        do_logout('recv end:' + op_key + ' ' + readyState + ' ' + status + ' ' + duration + 'ms');
        if (op_key == 'query_evt') {
            var v_time = new Date().getTime();
            if (v_time - m_agi_url.get_query_evt_tick() > 2000 || v_time < m_agi_url.get_query_evt_tick())//或者时间错误
            {
                start_query_event();
            }
            else {//定时器延迟3秒再启动query evt  
                window.setTimeout(function (self_obj) { self_obj.on_query_evt_timeout() }, 2000, self);
            }
        }
        else if (op_key == 'auth' && (m_auth_key.length <= 0))//授权请求失败,应该重试
        {
            if (self.get_retry_delay() > 0) {
                if (self.get_retry_delay() < 10)//可以认为无延迟,立即连接
                {
                    self.query_auth_key();
                }
                else {
                    m_retry_timer = window.setTimeout(function (self_obj) { self_obj.on_query_auth_timeout() }, self.get_retry_delay(), self);
                }
            }
            else {
                do_logout("ignore query auth retry_delay=0");
            }
        }
        else if (op_key == 'bind' && (readyState != 4 || status != 200))//执行绑定请求异常
        {
            if (self.get_retry_delay() > 0) {
                if (self.get_retry_delay() < 10)//可以认为无延迟,立即连接
                {
                    self.bind_account_try();
                }
                else {
                    m_retry_timer = window.setTimeout(function (self_obj) { self_obj.on_bind_timeout() }, self.get_retry_delay(), self);
                }
            } else {
                do_logout("ignore bind retry_delay=0");
            }
        }
    }
    //===
    this.stop_query_auth_timer = function () {
        if (m_retry_timer != 0) {
            window.clearTimeout(m_retry_timer);
            m_retry_timer = 0;
        }
    }
    this.on_bind_timeout = function () {
        self.bind_account_try();
    }
    this.on_query_auth_timeout = function () {
        self.query_auth_key();
    }

    this.on_query_evt_timeout = function () {
        start_query_event();
    }

    this.do_recv_message = function (js_Obj) {
        var cmd = (js_Obj != undefined) ? js_Obj.cmd : '';
        if (cmd == 'reply') {
            self.do_reply_msg(js_Obj);
        }
        else if (cmd == 'notify') {
            do_logout("->notify cmd");
            self.on_notify_msg(js_Obj);
        }
        else {
            do_logout("->ignore cmd:" + cmd);
        }
    }

    //立即回复请求的操作
    this.do_reply_msg = function (js_Obj) {
        var act = js_Obj.content.action;
        var act_id = js_Obj.content.act_id;
        var err_id = js_Obj.err_id;
        do_logout("->reply cmd:" + act);
        if (err_id == 0) {
            if (act == 'auth') self.do_auth_successed(js_Obj.content);
            else if (act == 'bind') self.do_bind_successed(js_Obj.content.acc_list);
            else if (act == 'unbind') self.do_unbind_successed(js_Obj.content.acc_list);
            else if (act == 'post_msg') self.do_post_msg_successed(js_Obj.content);
            else if (act == 'query_evt') self.do_query_evt_successed(js_Obj);
            else if (act == 'post_act') self.do_post_action_successed(js_Obj.content);
        }
        else {
            if (act == 'auth') self.on_auth_failed(m_user_key, err_id, js_Obj.err_desc);
            else if (act == 'bind') self.do_bind_failed(err_id, js_Obj.err_desc);
            else if (act == 'unbind') self.do_unbind_failed(err_id, js_Obj.err_desc);
            else if (act == 'post_msg') self.do_post_msg_failed(js_Obj.content);
            else if (act == 'query_evt') self.do_query_evt_failed(err_id, js_Obj.err_desc, js_Obj.content);
            else if (act == 'post_act') self.do_post_action_failed(js_Obj.content);

            if (err_id == '80000015' || js_Obj.err_desc == 'invalid auth key') {
                self.set_unbind_all_account();//授权错误，全部下线重新获取后上线
                self.query_auth_key();//重新获取授权，服务器要注意不要出现刚申请就错误的,授权获取成功bind后又不合法
            }
        }
    }

    this.do_query_evt_successed = function (js_Obj) {
        //alert(JSON.stringify(js_Obj));
        //id小于等于原来的就丢弃,可能重复了 
        var evt_array = js_Obj['content']['event'];
        if (u_isArray(evt_array)) {
            for (n = 0; n < evt_array.length; n++) {
                var event = evt_array[n];
                //保存最新收到的事件id,下次请求时提交上去
                if (typeof (event['oid']) == "object") {
                    var eid = parseInt(event.oid.eid);
                    if (eid > m_evt_eid)//小于的就是重复的丢弃
                    {
                        if ((m_evt_uid) > 0 && parseInt(event.oid.uid) > m_evt_uid + 1) {
                            do_logout("***lost uid pkt:" + m_evt_uid + '/' + event.oid.uid);
                        }
                        m_evt_uid = parseInt(event.oid.uid);
                        m_evt_eid = parseInt(eid);
                        this.do_event_msg(event);
                    }
                    else {
                        do_logout('**ignore event eid:' + eid + '/' + m_evt_eid + '->' + JSON.stringify(event));
                    }
                }
                else {
                    do_logout('***error oid param');
                }
            }
            //只要获取事件成功了处理后在onclose/do_recv_end立即再次获取
            m_agi_url.reset_query_evt_tick();
        }
        else {
            do_logout('***event is not array ***');
        }
    }

    this.do_query_evt_failed = function (err_id, err_desc, content) {
        do_logout('query evt failed :' + err_id + '->' + err_desc + '->' + JSON.stringify(content));
        //如果是等待超时的，就立即重试
        if (err_id == 80000103 || err_desc == 'timeout') {
        }
        else {//如果间隔足够长可以立即发起,建议可以3000ms以上就可以立即重新发起，让closed后判断m_query_evt_time 大于0
        }
    }

    this.do_event_msg = function (js_event) {
        this.on_recv_event(js_event);
    }

    this.do_auth_successed = function (content) {
        m_evt_uid = 0;
        m_evt_eid = 0;
        self.set_auth_key(content.auth_key);
        self.on_auth_successed(self.get_user_key(), self.get_auth_key());
        self.bind_account_try();
    }

    this.do_bind_failed = function (err_id, err_desc) {
        self.on_bind_failed('', err_id, err_desc);
    }

    this.do_bind_successed = function (acc_array) {
        for (n = 0; n < acc_array.length; n++) {
            var v_acc = acc_array[n].account;
            if (acc_array[n].err_id == 0) {
                m_enable_query_evt = true;
                var ep_acc = acc_array[n].param.acc_info.acc_data.account;
                m_bind_acc[ep_acc]=acc_array[n].param.acc_info.acc_data;
                do_logout('bind ep_acc success:' + ep_acc);
                self.on_bind_successed(ep_acc, acc_array[n].param);

                if (m_unbind_acc[ep_acc]) { 
                    delete m_unbind_acc[ep_acc];
                }
                if (m_unbind_acc[v_acc]) { 
                    delete m_unbind_acc[v_acc];
                }
                /*
                if (m_unbind_acc[v_acc]) {                    
                    do_logout('bind success:' + v_acc);
                    m_bind_acc[v_acc] = m_unbind_acc[v_acc];                   
                    delete m_unbind_acc[v_acc];
                    self.on_bind_successed(v_acc, acc_array[n].param);
                }
                else {
                    do_logout('ignore bind success:' + v_acc);
                }
                */
            }
            else {
                do_logout('**bind failed:' + acc_array[n].err_id + '[' + v_acc + ']');
                self.on_bind_failed(v_acc, acc_array[n].err_id, acc_array[n].err_desc);
            }
            start_query_event();
        }
    }
    this.do_unbind_failed = function (err_id, err_desc) {
        self.on_unbind_failed('', err_id, err_desc);
    }
    this.do_unbind_successed = function (acc_array) {
        for (n = 0; n < acc_array.length; n++) {
            var v_acc = acc_array[n].account;
            if (acc_array[n].err_id == 0) {
                if (m_bind_acc[v_acc]) {
                    delete m_bind_acc[v_acc];
                    do_logout('unbind delete:' + v_acc);
                }
                self.on_unbind_successed(v_acc, acc_array[n].param);
            }
            else {
                self.on_unbind_failed(v_acc, acc_array[n].err_id, acc_array[n].err_desc);
            }
        }
    }

    this.do_post_msg_successed = function (js_content) { this.on_post_msg_successed(js_content); }
    this.do_post_msg_failed = function (js_content) { this.on_post_msg_failed(js_content); }
    this.do_post_action_successed = function (js_content) { this.on_post_action_successed(js_content); }
    this.do_post_action_failed = function (js_content) { this.on_post_action_failed(js_content); }

    var do_logout = function (s) {
        if (!m_logout) return;
        else u_logout(s);
    }
    this.d_logout = function (s) {
        if (!m_logout) return;
        else u_logout(s);
    }
    //--虚函数
    this.on_notify_msg = function (js_Obj) { }
    this.on_recv_text = function (data) { }
    this.is_open = function () { return m_agi_url.is_open(); }
    this.send = function (data, timeout, op_key) { return m_agi_url.send(data, timeout, op_key); }
    //---
    this.on_bind_wait_timeout = function (t_ms) { do_logout('*bind acc timeout:' + t_ms + 'ms'); }
    this.on_unbind_successed = function (v_acc, param) { do_logout('unbind successed:' + v_acc); }
    this.on_unbind_failed = function (v_acc, err_id, err_desc) { do_logout('unbind failed:' + v_acc); }
    this.on_bind_failed = function (v_acc, err_id, err_desc) { do_logout('bind failed:' + v_acc); }
    this.on_bind_successed = function (v_acc, param) { do_logout('bind successed:' + v_acc); }
    this.on_unbind_all_account = function (acc_array) { do_logout('unbind all'); }
    this.on_set_unbind_all_account = function (acc_array) { do_logout('unbind set all'); }
    //--
    this.on_connected = function (url) { do_logout('url connected:' + url); }
    this.on_disconnected = function (url) { do_logout('url disconnected:' + url); }
    this.on_url_error = function (url) { do_logout('url error:' + url); }
    this.on_recv_event = function (json_evt) { do_logout('event->' + JSON.stringify(json_evt)); }
    this.on_post_msg_successed = function (js_content) { do_logout('post msg successed:' + JSON.stringify(js_content)); }
    this.on_post_msg_failed = function (js_content) { do_logout('**post msg failed:' + JSON.stringify(js_content)); }
    this.on_post_action_successed = function (js_content) { do_logout('post action successed:' + JSON.stringify(js_content)); }
    this.on_post_action_failed = function (js_content) { do_logout('**post action failed:' + JSON.stringify(js_content)); }
    //----默认使用第一个绑定的账号
    //多个时需要指定分机号
    var make_bind_acc = function (acc) {
        if (u_isEmpty(acc)) {
            var v_key = Object.keys(m_bind_acc);
            if (u_isNotEmpty(v_key)) { acc = v_key[0]; }
            else { acc = ''; }
        }
        else {
            if (!m_bind_acc.hasOwnProperty(acc)) { return ''; }
        }
        return acc;
    }
    //==================
    var do_post_action = function (json_post) {
        var v_acc = make_bind_acc(json_post['account']);
        if (u_isEmpty(v_acc)) { return ai_err_invalid_param; }
        json_post['account'] = v_acc; //如果为空强制修改第一个绑定的账号
        return do_action_param('post_act', json_post);
    }

    var do_post_action_param = function (acc, act, json_param) {
        var json_post = {
            'account': acc,
            'action': act,
            'param': json_param
        }
        return do_post_action(json_post);
    }

    this.call_out = function (acc, code, dst_to, au_dev) {
        var json_param = {
            'cmd': 'start',
            'code': u_data(code),
            'dst_to': u_data(dst_to),
            'au_dev': u_data(au_dev)
        }
        return do_post_action_param(acc, 'call_out', json_param);
    }
    this.hangup = function (acc, dlg_id) {
        var json_param = {
            'dlg_id': u_data(dlg_id)
        }
        return do_post_action_param(acc, 'hangup', json_param);
    }
    //usb设备才有这种操作，为兼容sip电话操作,外部可以不要用这种特有的设备操作
    this.offhook = function (acc) {
        var json_param = {};
        return do_post_action_param(acc, 'offhook', json_param);
    }
    this.answer = function (acc, dlg_id) {
        var json_param = {
            'dlg_id': u_data(dlg_id)
        }
        return do_post_action_param(acc, 'answer', json_param);
    }
    this.playfile_start = function (acc, file, repeat) {
        var json_param = {
            'cmd': 'start',
            'file': u_data(file),
            'repeat': u_data(repeat)
        }
        return do_post_action_param(acc, 'play_file', json_param);
    }
    this.playfile_stop = function (acc, handle) {
        var json_param = {
            'cmd': 'stop',
            'handle': u_data(handle)
        }
        return do_post_action_param(acc, 'play_file', json_param);
    }
    this.dial_dtmf = function (acc, dtmf) {
        var json_param = {
            'cmd': 'start',
            'code': u_data(dtmf)
        }
        return do_post_action_param(acc, 'dial_dtmf', json_param);
    }
    this.do_ctrl = function (acc, ctrl, v) {
        var json_param = {};
        json_param[ctrl] = u_data(v);
        return do_post_action_param(acc, 'do_ctrl', json_param);
    }
    //---
}

//---------------------------------
function u_is_err_not_connected(e) {
    if (e == ai_err_not_open) return true;
    else return false;
}
function u_isNotEmpty(res) {
    if (res != undefined && res != "" && res.length > 0)
        return true;
    return false;
}
function u_isEmpty(res) {
    if (res == undefined || res == "" || res.length <= 0)
        return true;
    else return false;
}
function u_data(res) {
    if (u_isEmpty(res)) return '';
    else return res;
}

function u_setCookie() {
    var d = new Date();
    d.setTime(d.getTime() + (exdays * 24 * 60 * 60 * 1000));
    var expires = "expires=" + d.toGMTString();
    document.cookie = expires;
}

function u_Format2d(v) {
    if (v < 10) return '0' + v.toString();
    else return v.toString();
}

function u_Format3d(v) {
    if (v < 10) return '00' + v.toString();
    else if (v < 100) return '0' + v.toString();
    else return v.toString();
}

function u_GetNowTimeString() {
    var vd = new Date();
    return '[' + u_Format2d(vd.getHours()) + ':' + u_Format2d(vd.getMinutes()) + ':' + u_Format2d(vd.getSeconds()) + '.' + u_Format3d(vd.getMilliseconds()) + ']';
}

function u_getutc_ms() {
    return new Date().getTime();
}

function u_md5(s) {
    return s.MD5(32);
}
function u_isArray(a) {
    if (Object.prototype.toString.call(a) === '[Object Array]'
        || Array.isArray(a)) return true;
    else return false;
}

function u_url_protocol(url)//HTTP:// -> http
{
    var v_index = url.indexOf(':');
    if (v_index > 0) {
        return url.substr(0, v_index).toLowerCase();
    }
    else return "";
}

function u_logout(s) {
    var v = u_GetNowTimeString() + ' ' + s;
    console.log(v);
}

function u_gbk2utf8(text) {
    var encode = encodeURIComponent(text);
    return decodeURIComponent(encode);
    //var utf8text = unescape(decoded);
    //return utf8text;
}

// 给字符串对象增加一个MD5的方法
String.prototype.MD5 = function (bit) {
    var sMessage = this;
    function RotateLeft(lValue, iShiftBits) { return (lValue << iShiftBits) | (lValue >>> (32 - iShiftBits)); }
    function AddUnsigned(lX, lY) {
        var lX4, lY4, lX8, lY8, lResult;
        lX8 = (lX & 0x80000000);
        lY8 = (lY & 0x80000000);
        lX4 = (lX & 0x40000000);
        lY4 = (lY & 0x40000000);
        lResult = (lX & 0x3FFFFFFF) + (lY & 0x3FFFFFFF);
        if (lX4 & lY4) return (lResult ^ 0x80000000 ^ lX8 ^ lY8);
        if (lX4 | lY4) {
            if (lResult & 0x40000000) return (lResult ^ 0xC0000000 ^ lX8 ^ lY8);
            else return (lResult ^ 0x40000000 ^ lX8 ^ lY8);
        } else return (lResult ^ lX8 ^ lY8);
    }
    function F(x, y, z) { return (x & y) | ((~x) & z); }
    function G(x, y, z) { return (x & z) | (y & (~z)); }
    function H(x, y, z) { return (x ^ y ^ z); }
    function I(x, y, z) { return (y ^ (x | (~z))); }
    function FF(a, b, c, d, x, s, ac) {
        a = AddUnsigned(a, AddUnsigned(AddUnsigned(F(b, c, d), x), ac));
        return AddUnsigned(RotateLeft(a, s), b);
    }
    function GG(a, b, c, d, x, s, ac) {
        a = AddUnsigned(a, AddUnsigned(AddUnsigned(G(b, c, d), x), ac));
        return AddUnsigned(RotateLeft(a, s), b);
    }
    function HH(a, b, c, d, x, s, ac) {
        a = AddUnsigned(a, AddUnsigned(AddUnsigned(H(b, c, d), x), ac));
        return AddUnsigned(RotateLeft(a, s), b);
    }
    function II(a, b, c, d, x, s, ac) {
        a = AddUnsigned(a, AddUnsigned(AddUnsigned(I(b, c, d), x), ac));
        return AddUnsigned(RotateLeft(a, s), b);
    }
    function ConvertToWordArray(sMessage) {
        var lWordCount;
        var lMessageLength = sMessage.length;
        var lNumberOfWords_temp1 = lMessageLength + 8;
        var lNumberOfWords_temp2 = (lNumberOfWords_temp1 - (lNumberOfWords_temp1 % 64)) / 64;
        var lNumberOfWords = (lNumberOfWords_temp2 + 1) * 16;
        var lWordArray = Array(lNumberOfWords - 1);
        var lBytePosition = 0;
        var lByteCount = 0;
        while (lByteCount < lMessageLength) {
            lWordCount = (lByteCount - (lByteCount % 4)) / 4;
            lBytePosition = (lByteCount % 4) * 8;
            lWordArray[lWordCount] = (lWordArray[lWordCount] | (sMessage.charCodeAt(lByteCount) << lBytePosition));
            lByteCount++;
        }
        lWordCount = (lByteCount - (lByteCount % 4)) / 4;
        lBytePosition = (lByteCount % 4) * 8;
        lWordArray[lWordCount] = lWordArray[lWordCount] | (0x80 << lBytePosition);
        lWordArray[lNumberOfWords - 2] = lMessageLength << 3;
        lWordArray[lNumberOfWords - 1] = lMessageLength >>> 29;
        return lWordArray;
    }
    function WordToHex(lValue) {
        var WordToHexValue = "", WordToHexValue_temp = "", lByte, lCount;
        for (lCount = 0; lCount <= 3; lCount++) {
            lByte = (lValue >>> (lCount * 8)) & 255;
            WordToHexValue_temp = "0" + lByte.toString(16);
            WordToHexValue = WordToHexValue + WordToHexValue_temp.substr(WordToHexValue_temp.length - 2, 2);
        }
        return WordToHexValue;
    }
    var x = Array();
    var k, AA, BB, CC, DD, a, b, c, d
    var S11 = 7, S12 = 12, S13 = 17, S14 = 22;
    var S21 = 5, S22 = 9, S23 = 14, S24 = 20;
    var S31 = 4, S32 = 11, S33 = 16, S34 = 23;
    var S41 = 6, S42 = 10, S43 = 15, S44 = 21;
    // Steps 1 and 2. Append padding bits and length and convert to words 
    x = ConvertToWordArray(sMessage);
    // Step 3. Initialise 
    a = 0x67452301; b = 0xEFCDAB89; c = 0x98BADCFE; d = 0x10325476;
    // Step 4. Process the message in 16-word blocks 
    for (k = 0; k < x.length; k += 16) {
        AA = a; BB = b; CC = c; DD = d;
        a = FF(a, b, c, d, x[k + 0], S11, 0xD76AA478);
        d = FF(d, a, b, c, x[k + 1], S12, 0xE8C7B756);
        c = FF(c, d, a, b, x[k + 2], S13, 0x242070DB);
        b = FF(b, c, d, a, x[k + 3], S14, 0xC1BDCEEE);
        a = FF(a, b, c, d, x[k + 4], S11, 0xF57C0FAF);
        d = FF(d, a, b, c, x[k + 5], S12, 0x4787C62A);
        c = FF(c, d, a, b, x[k + 6], S13, 0xA8304613);
        b = FF(b, c, d, a, x[k + 7], S14, 0xFD469501);
        a = FF(a, b, c, d, x[k + 8], S11, 0x698098D8);
        d = FF(d, a, b, c, x[k + 9], S12, 0x8B44F7AF);
        c = FF(c, d, a, b, x[k + 10], S13, 0xFFFF5BB1);
        b = FF(b, c, d, a, x[k + 11], S14, 0x895CD7BE);
        a = FF(a, b, c, d, x[k + 12], S11, 0x6B901122);
        d = FF(d, a, b, c, x[k + 13], S12, 0xFD987193);
        c = FF(c, d, a, b, x[k + 14], S13, 0xA679438E);
        b = FF(b, c, d, a, x[k + 15], S14, 0x49B40821);
        a = GG(a, b, c, d, x[k + 1], S21, 0xF61E2562);
        d = GG(d, a, b, c, x[k + 6], S22, 0xC040B340);
        c = GG(c, d, a, b, x[k + 11], S23, 0x265E5A51);
        b = GG(b, c, d, a, x[k + 0], S24, 0xE9B6C7AA);
        a = GG(a, b, c, d, x[k + 5], S21, 0xD62F105D);
        d = GG(d, a, b, c, x[k + 10], S22, 0x2441453);
        c = GG(c, d, a, b, x[k + 15], S23, 0xD8A1E681);
        b = GG(b, c, d, a, x[k + 4], S24, 0xE7D3FBC8);
        a = GG(a, b, c, d, x[k + 9], S21, 0x21E1CDE6);
        d = GG(d, a, b, c, x[k + 14], S22, 0xC33707D6);
        c = GG(c, d, a, b, x[k + 3], S23, 0xF4D50D87);
        b = GG(b, c, d, a, x[k + 8], S24, 0x455A14ED);
        a = GG(a, b, c, d, x[k + 13], S21, 0xA9E3E905);
        d = GG(d, a, b, c, x[k + 2], S22, 0xFCEFA3F8);
        c = GG(c, d, a, b, x[k + 7], S23, 0x676F02D9);
        b = GG(b, c, d, a, x[k + 12], S24, 0x8D2A4C8A);
        a = HH(a, b, c, d, x[k + 5], S31, 0xFFFA3942);
        d = HH(d, a, b, c, x[k + 8], S32, 0x8771F681);
        c = HH(c, d, a, b, x[k + 11], S33, 0x6D9D6122);
        b = HH(b, c, d, a, x[k + 14], S34, 0xFDE5380C);
        a = HH(a, b, c, d, x[k + 1], S31, 0xA4BEEA44);
        d = HH(d, a, b, c, x[k + 4], S32, 0x4BDECFA9);
        c = HH(c, d, a, b, x[k + 7], S33, 0xF6BB4B60);
        b = HH(b, c, d, a, x[k + 10], S34, 0xBEBFBC70);
        a = HH(a, b, c, d, x[k + 13], S31, 0x289B7EC6);
        d = HH(d, a, b, c, x[k + 0], S32, 0xEAA127FA);
        c = HH(c, d, a, b, x[k + 3], S33, 0xD4EF3085);
        b = HH(b, c, d, a, x[k + 6], S34, 0x4881D05);
        a = HH(a, b, c, d, x[k + 9], S31, 0xD9D4D039);
        d = HH(d, a, b, c, x[k + 12], S32, 0xE6DB99E5);
        c = HH(c, d, a, b, x[k + 15], S33, 0x1FA27CF8);
        b = HH(b, c, d, a, x[k + 2], S34, 0xC4AC5665);
        a = II(a, b, c, d, x[k + 0], S41, 0xF4292244);
        d = II(d, a, b, c, x[k + 7], S42, 0x432AFF97);
        c = II(c, d, a, b, x[k + 14], S43, 0xAB9423A7);
        b = II(b, c, d, a, x[k + 5], S44, 0xFC93A039);
        a = II(a, b, c, d, x[k + 12], S41, 0x655B59C3);
        d = II(d, a, b, c, x[k + 3], S42, 0x8F0CCC92);
        c = II(c, d, a, b, x[k + 10], S43, 0xFFEFF47D);
        b = II(b, c, d, a, x[k + 1], S44, 0x85845DD1);
        a = II(a, b, c, d, x[k + 8], S41, 0x6FA87E4F);
        d = II(d, a, b, c, x[k + 15], S42, 0xFE2CE6E0);
        c = II(c, d, a, b, x[k + 6], S43, 0xA3014314);
        b = II(b, c, d, a, x[k + 13], S44, 0x4E0811A1);
        a = II(a, b, c, d, x[k + 4], S41, 0xF7537E82);
        d = II(d, a, b, c, x[k + 11], S42, 0xBD3AF235);
        c = II(c, d, a, b, x[k + 2], S43, 0x2AD7D2BB);
        b = II(b, c, d, a, x[k + 9], S44, 0xEB86D391);
        a = AddUnsigned(a, AA); b = AddUnsigned(b, BB); c = AddUnsigned(c, CC); d = AddUnsigned(d, DD);
    }
    if (bit == 32) {
        return WordToHex(a) + WordToHex(b) + WordToHex(c) + WordToHex(d);
    }
    else {
        return WordToHex(b) + WordToHex(c);
    }
}


function xhr_post_json(url, timeout, op_key) {
    var m_timeout = timeout;
    var m_url = url;
    var m_op_key = op_key;
    var m_begin_time = new Date().getTime();
    var self_xhr = this;
    var create_XHR = function () {
        var XHR = [  //兼容不同浏览器和版本得创建函数数组
            function () { return new XMLHttpRequest() },
            function () { return new ActiveXObject("Msxml2.XMLHTTP") },
            function () { return new ActiveXObject("Msxml3.XMLHTTP") },
            function () { return new ActiveXObject("Microsoft.XMLHTTP") }
        ];
        var xhr_ = null;
        for (var i = 0; i < XHR.length; i++) {
            try {
                xhr_ = XHR[i]();
            } catch (e) {
                continue;
            }
            break;
        }
        return xhr_;
    }
    /*https://developer.mozilla.org/zh-CN/docs/Web/API/XMLHttpRequest/upload
    onloadstart	获取开始
    onprogress	数据传输进行中
    onabort	获取操作终止
    onerror	获取失败
    onload	获取成功
    ontimeout	获取操作在用户规定的时间内未完成
    onloadend	获取完成（不论成功与否）
    */
    //在调用 abort() 方法前，应先清除 onreadystatechange 事件处理函数，因为 IE 和 Mozilla 在请求中止后也会激活这个事件处理函数。如果给 onreadystatechange 属性设置为 null，则 IE 会发生异常，所以为它设置一个空函数。
    //xhr.onreadystatechange = function () {};  //清理事件响应函数
    //xhr.abort();  //中止请求
    //XMLHttpRequest 对象通过 responseText、responseBody、responseStream 或 responseXML 属性获取响应信息，说明如下表所示，它们都是只读属性。
    /*
    responseBody	将响应信息正文以 Unsigned Byte 数组形式返回
    responseStream 	以 ADO Stream 对象的形式返回响应信息
    responseText	将响应信息作为字符串返回
    responseXML	    将响应信息格式化为 XML 文档格式返回

    //readyState
    0	未初始化。表示对象已经建立，但是尚未初始化，尚未调用 open() 方法
    1	初始化。表示对象已经建立，尚未调用 send() 方法
    2	发送数据。表示 send() 方法已经调用，但是当前的状态及 HTTP 头未知
    3	数据传送中。已经接收部分数据，因为响应及 HTTP 头不安全，这时通过 responseBody 和 responseText 获取部分数据会出现错误
    4	完成。数据接收完毕，此时可以通过 responseBody 和 responseText 获取完整的响应数据
    */
    this.send = function (s_data) {
        var xhr = create_XHR();
        xhr.open('POST', m_url, true);
        xhr.timeout = m_timeout;
        xhr.withCredentials = false;
        xhr.setRequestHeader("Content-Type", "application/json");
        xhr.ontimeout = function (e) {
            self_xhr.ontimeout();
        }
        xhr.onerror = function () {
            self_xhr.onerror(0);
        }
        xhr.onload = function () {
            //u_logout('onload');
        }
        xhr.onloadstart = function () {
            u_logout('onloadstart [' + m_op_key + ']' + ':' + m_url);
        }
        xhr.onprogress = function () {
            //u_logout('onprogress');
        }
        xhr.onabort = function () {
            u_logout('onabort');
        }
        xhr.onloadend = function () {
            self_xhr.onclose(m_op_key, xhr.readyState, xhr.status, new Date().getTime() - m_begin_time);
        }
        xhr.onreadystatechange = function () {
            if (xhr.readyState == 4) {
                if (xhr.status == 200) {
                    self_xhr.onmessage((xhr.responseText));
                }
                else if (xhr.status > 0) {
                    self_xhr.onerror(xhr.status);
                }
            }
        }
        xhr.send(s_data);
        return xhr;
    }
    this.send_json = function (json_data) {
        this.send(JSON.stringify(json_data));
    }
    //---
    this.ontimeout = function () { u_logout('xhr timeout'); }
    this.onopen = function () { },
        this.onmessage = function (s_data) { u_logout('xhr msg:' + s_data.length); },
        this.onerror = function () { u_logout('xhr error'); },
        this.onclose = function () { u_logout('xhr onclose...'); }
}
